/**
 * Copyright 2006 Joseph M. Ferner
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *    
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.fernsroth.easyio;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import com.fernsroth.easyio.util.EasyIOUtils;

/**
 * 
 * @author Joseph M. Ferner (Near Infinity Corporation)
 */
public class EasyIOFormatter {

    /**
     * the default number of bytes per lines.
     */
    private static final int DEFAULT_BYTES_PER_LINE = 16;

    /**
     * the default number of characters in the offset.
     */
    private static final int DEFAULT_CHARACTERS_IN_OFFSET = 8;

    /**
     * the input stream to read from.
     */
    private InputStream inputStream;

    /**
     * the output stream to output to.
     */
    private OutputStream outputStream;

    /**
     * number of bytes per line.
     */
    private int bytesPerLine = DEFAULT_BYTES_PER_LINE;

    /**
     * the number of characters in the offset.
     */
    private int charactersInOffset = DEFAULT_CHARACTERS_IN_OFFSET;

    /**
     * print a pretty string of the hex.
     * @param in the input stream.
     * @return the pretty string.
     * @throws IOException 
     */
    public static String print(InputStream in) throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        print(in, baos);
        return baos.toString();
    }

    /**
     * print a pretty string of the hex.
     * @param in the input stream.
     * @param out the output stream.
     * @throws IOException 
     */
    private static void print(InputStream in, OutputStream out)
            throws IOException {
        EasyIOFormatter formatter = new EasyIOFormatter();
        formatter.setInputStream(in);
        formatter.setOutputStream(out);
        formatter.print();
    }

    /**
     * prints data.
     * @param data the data to print.
     * @return the string.
     */
    public static String print(byte[] data) {
        return print(data, 0, data.length);
    }

    /**
     * prints data.
     * @param data the data to print.
     * @param offset the offset into the byte array.
     * @param length the length of the byte array.
     * @return the string.
     */
    public static String print(byte[] data, int offset, int length) {
        try {
            return print(new ByteArrayInputStream(data, offset, length));
        } catch (IOException e) {
            throw new RuntimeException("unexpected IOException", e);
        }
    }

    /**
     * prints.
     * @throws IOException 
     */
    public void print() throws IOException {
        int offset = 0;
        byte[] buffer = new byte[getBytesPerLine()];
        while (true) {
            int read = this.inputStream.read(buffer);
            if (read == -1) {
                break;
            }
            printOffset(offset);
            this.outputStream.write("  ".getBytes());
            printHex(buffer, read);
            this.outputStream.write("  ".getBytes());
            printASCII(buffer, read);
            this.outputStream.write('\n');
            offset += buffer.length;
        }
    }

    /**
     * prints the ASCII values.
     * @param buffer the buffer to print.
     * @param read the number of valid entries in the buffer.
     * @throws IOException 
     */
    private void printASCII(byte[] buffer, int read) throws IOException {
        int i;
        for (i = 0; i < read; i++) {
            if (isPrintable(buffer[i])) {
                this.outputStream.write(buffer[i]);
            } else {
                this.outputStream.write('.');
            }
        }
        for (; i < getBytesPerLine(); i++) {
            this.outputStream.write(' ');
        }
    }

    /**
     * checks if a value is printable.
     * @param b the value to check.
     * @return true, if printable. false, if not.
     */
    private boolean isPrintable(byte b) {
        if (b >= ' ' && b <= '~') {
            return true;
        }
        return false;
    }

    /**
     * prints the hex values.
     * @param buffer the buffer to print.
     * @param read the number of valid entries in the buffer.
     * @throws IOException 
     */
    private void printHex(byte[] buffer, int read) throws IOException {
        int i;
        for (i = 0; i < read; i++) {
            this.outputStream.write(EasyIOUtils.toHexString(buffer[i], 2)
                    .getBytes());
            this.outputStream.write(' ');
        }
        for (; i < getBytesPerLine(); i++) {
            this.outputStream.write("   ".getBytes());
        }
    }

    /**
     * @param offset the offset to print.
     * @throws IOException 
     */
    private void printOffset(int offset) throws IOException {
        this.outputStream.write(EasyIOUtils.toHexString(offset,
                getCharactersInOffset()).getBytes());
    }

    /**
     * @param inputStream the inputStream to set
     */
    public void setInputStream(InputStream inputStream) {
        this.inputStream = inputStream;
    }

    /**
     * @param outputStream the outputStream to set
     */
    public void setOutputStream(OutputStream outputStream) {
        this.outputStream = outputStream;
    }

    /**
     * @return the bytesPerLine
     */
    public int getBytesPerLine() {
        return this.bytesPerLine;
    }

    /**
     * @param bytesPerLine the bytesPerLine to set
     */
    public void setBytesPerLine(int bytesPerLine) {
        this.bytesPerLine = bytesPerLine;
    }

    /**
     * @return the charactersInOffset
     */
    public int getCharactersInOffset() {
        return this.charactersInOffset;
    }

    /**
     * @param charactersInOffset the charactersInOffset to set
     */
    public void setCharactersInOffset(int charactersInOffset) {
        this.charactersInOffset = charactersInOffset;
    }

}
