/**
 * Copyright 2006 Joseph M. Ferner
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *    
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.fernsroth.easyio;

import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.math.BigInteger;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import com.fernsroth.easyio.exception.EasyIOException;
import com.fernsroth.easyio.util.BeanUtils;
import com.fernsroth.easyio.util.EasyIOUtils;

/**
 * input stream for reading and writing binary data.
 * @author Joseph M. Ferner (Near Infinity Corporation)
 */
public class EasyIOInputStream extends FilterInputStream implements
        IInputStream {

    /**
     * the currect working byte.
     */
    private int currentByte;

    /**
     * number of bits left in {@link currentByte}.
     */
    private int bitsLeft;

    /**
     * the field handler registry.
     */
    private FieldHandlerRegistry fieldHandlerRegistry = DefaultFieldHandlerRegistry
            .getInstance();

    /**
     * constructor.
     * @param in the input stream to wrap.
     */
    public EasyIOInputStream(InputStream in) {
        super(in);
        this.currentByte = 0;
        this.bitsLeft = 0;
    }

    /**
     * constructor.
     * @param in the input stream.
     */
    public EasyIOInputStream(IInputStream in) {
        this(new InputStreamAdapter(in));
    }

    /**
     * reads in a object.
     * @param <T> the object type.
     * @param obj the object to read data into.
     * @return the read in object.
     * @throws EasyIOException 
     * @throws IOException 
     */
    public <T> T read(T obj) throws EasyIOException, IOException {
        List<Field> fields = BeanUtils.getAllFields(obj.getClass());
        Queue<Field> fieldsQueue = new LinkedList<Field>(fields);

        while (!fieldsQueue.isEmpty()) {
            Field field = fieldsQueue.peek();
            FieldHandler handler = this.fieldHandlerRegistry.lookup(obj, field);
            if (handler != null) {
                handler.read(this, obj, fieldsQueue);
            } else {
                fieldsQueue.remove();
            }
        }

        return obj;
    }

    /**
     * reads a little-endien int64.
     * @return the int.
     * @throws IOException 
     */
    public long readINT64() throws IOException {
        byte[] data = new byte[8];
        if (read(data) != data.length) {
            throw new IOException("read past end of stream");
        }
        long l1 = (((long) data[0] & 0xff) << 0)
                | (((long) data[1] & 0xff) << 8)
                | (((long) data[2] & 0xff) << 16)
                | (((long) data[3] & 0xff) << 24);
        long l2 = (((long) data[4] & 0xff) << 0)
                | (((long) data[5] & 0xff) << 8)
                | (((long) data[6] & 0xff) << 16)
                | (((long) data[7] & 0xff) << 24);
        return (l1 << 0) | (l2 << 32);
    }

    /**
     * reads a little-endien uint64.
     * @return the int.
     * @throws IOException 
     * @throws IOException 
     */
    public BigInteger readUINT64() throws IOException {
        byte[] data = new byte[8];
        if (read(data) != data.length) {
            throw new IOException("read past end of stream");
        }
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < data.length; i++) {
            sb.append(EasyIOUtils.toHexString(data[i], 2));
        }
        return new BigInteger(sb.toString(), 16);
    }

    /**
     * reads a little-endien uint32.
     * @return the int.
     * @throws IOException 
     */
    public long readUINT32() throws IOException {
        byte[] data = new byte[4];
        if (read(data) != data.length) {
            throw new IOException("read past end of stream");
        }
        return (((long) data[0] & 0xff) << 0) | (((long) data[1] & 0xff) << 8)
                | (((long) data[2] & 0xff) << 16)
                | (((long) data[3] & 0xff) << 24);
    }

    /**
     * reads a little-endien uint16.
     * @return the int.
     * @throws IOException 
     */
    public int readUINT16() throws IOException {
        byte[] data = new byte[2];
        if (read(data) != data.length) {
            throw new IOException("read past end of stream");
        }
        return (((int) data[0] & 0xff) << 0) | (((int) data[1] & 0xff) << 8);
    }

    /**
     * reads a uint8.
     * @return the int.
     * @throws IOException 
     */
    public int readUINT8() throws IOException {
        byte[] data = new byte[1];
        if (read(data) != data.length) {
            throw new IOException("read past end of stream");
        }
        return (int) data[0] & 0xff;
    }

    /**
     * read the specified number of bits from the input stream.
     * @param buffer the data array to read it into.
     * @param bits the number of bits to read.
     * @return the number of bits read.
     * @throws IOException 
     */
    public int readBits(int bits, byte[] buffer) throws IOException {
        return readBits(bits, buffer, 0, buffer.length);
    }

    /**
     * read the specified number of bits from the input stream
     * (the bits will be left aligned within the buffer).
     * @param buffer the data array to read it into.
     * @param bits the number of bits to read.
     * @param offset offset into buffer.
     * @param length length of buffer.
     * @return the number of bits read.
     * @throws IOException 
     */
    public int readBits(int bits, byte[] buffer, int offset, int length)
            throws IOException {
        int bitsLeftToRead = bits;
        int bufferIdx = offset;
        int bitsRead = 0;
        while (bitsLeftToRead > 0) {
            int bitsToRead = Math.min(8, bitsLeftToRead);
            int data = readBits(bitsToRead);
            bitsRead += bitsToRead;
            buffer[bufferIdx++] = (byte) data;
            bitsLeftToRead -= bitsToRead;
        }
        return bitsRead;
    }

    /**
     * bit masks for {@link EasyIOInputStream#readBits(int)}.
     */
    private static final int[] BIT_MASKS = new int[] { 0x01, 0x03, 0x07, 0x0f,
            0x1f, 0x3f, 0x7f, 0xff };

    /**
     * @param bits number of bits to read (<=8).
     * @return the bits read.
     * @throws IOException 
     */
    public int readBits(int bits) throws IOException {
        int ret = 0;
        int bitsToRead = bits;

        if (bits == 0) {
            return 0;
        }
        if (bits > 8) {
            throw new IOException("bits requested but be <= 8");
        }

        while (bitsToRead > 0) {
            if (this.bitsLeft == 0) {
                this.currentByte = super.read();
                if (this.currentByte == -1) {
                    return -1;
                }
                this.bitsLeft = 8;
            }

            int bitsRead = Math.min(bitsToRead, this.bitsLeft);
            ret |= this.currentByte >> (this.bitsLeft - bitsRead);
            bitsToRead -= bitsRead;
            this.bitsLeft -= bitsRead;
            ret <<= bitsToRead;
        }

        return ret & BIT_MASKS[bits - 1];
    }

}
