/**
 * Copyright 2006 Joseph M. Ferner
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *    
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.fernsroth.easyio;

import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.math.BigInteger;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import com.fernsroth.easyio.exception.EasyIOException;
import com.fernsroth.easyio.util.BeanUtils;

/**
 * @author Joseph M. Ferner (Near Infinity Corporation)
 */
public class EasyIOOutputStream extends FilterOutputStream {

    /**
     * count of bytes that have been processed.
     */
    private long count = 0;

    /**
     * the field handler registry.
     */
    private FieldHandlerRegistry fieldHandlerRegistry = DefaultFieldHandlerRegistry
            .getInstance();

    /**
     * constructor.
     * @param out the output stream to wrap.
     */
    public EasyIOOutputStream(OutputStream out) {
        super(out);
    }

    /**
     * constructor.
     * @param out the output stream to wrap.
     */
    public EasyIOOutputStream(IOutputStream out) {
        this(new OutputStreamAdapter(out));
    }

    /**
     * writes an object to the stream.
     * @param obj the object to write.
     * @throws EasyIOException 
     * @throws IOException 
     */
    public void write(Object obj) throws EasyIOException, IOException {
        List<Field> fields = BeanUtils.getAllFields(obj.getClass());
        Queue<Field> fieldsQueue = new LinkedList<Field>(fields);
        Field field;

        while (!fieldsQueue.isEmpty()) {
            field = fieldsQueue.peek();
            FieldHandler handler = this.fieldHandlerRegistry.lookup(obj, field);
            if (handler != null) {
                handler.write(this, obj, fieldsQueue);
            } else {
                fieldsQueue.remove();
            }
        }
    }

    /**
     * writes an 64-bit integer (little endien).
     * @param val the value to write.
     * @throws IOException 
     */
    public void writeINT64(long val) throws IOException {
        byte[] data = new byte[8];
        data[0] = (byte) ((val >> 0) & 0xff);
        data[1] = (byte) ((val >> 8) & 0xff);
        data[2] = (byte) ((val >> 16) & 0xff);
        data[3] = (byte) ((val >> 24) & 0xff);
        data[4] = (byte) ((val >> 32) & 0xff);
        data[5] = (byte) ((val >> 40) & 0xff);
        data[6] = (byte) ((val >> 48) & 0xff);
        data[7] = (byte) ((val >> 56) & 0xff);
        write(data);
    }

    /**
     * write a uint64.
     * @param val the value to write.
     * @throws IOException 
     */
    public void writeUINT64(BigInteger val) throws IOException {
        String valstr = val.toString(16);
        byte[] data = new byte[8];
        for (int i = 0; i < data.length; i++) {
            data[i] = (byte) Integer.parseInt(valstr.substring(i * 2,
                    (i + 1) * 2), 16);
        }
        write(data);
    }

    /**
     * writes an unsigned 64-bit integer (little endien).
     * @param val the value to write.
     * @throws IOException 
     */
    public void writeUINT32(long val) throws IOException {
        byte[] data = new byte[4];
        data[0] = (byte) ((val >> 0) & 0xff);
        data[1] = (byte) ((val >> 8) & 0xff);
        data[2] = (byte) ((val >> 16) & 0xff);
        data[3] = (byte) ((val >> 24) & 0xff);
        write(data);
    }

    /**
     * writes an unsigned 16-bit integer (little endien).
     * @param val the value to write.
     * @throws IOException 
     */
    public void writeUINT16(int val) throws IOException {
        byte[] data = new byte[2];
        data[0] = (byte) ((val >> 0) & 0xff);
        data[1] = (byte) ((val >> 8) & 0xff);
        write(data);
    }

    /**
     * writes an unsigned 8-bit integer.
     * @param val the value to write.
     * @throws IOException 
     */
    public void writeUINT8(int val) throws IOException {
        write(val);
    }

    /**
     * gets the count of bytes written.
     * @return the count of bytes written.
     */
    public long getCount() {
        return this.count;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void write(int b) throws IOException {
        super.write(b);
        this.count++;
    }
}
