/**
 * Copyright 2006 Joseph M. Ferner
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *    
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.fernsroth.easyio.field.handler;

import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Queue;

import com.fernsroth.easyio.EasyIOInputStream;
import com.fernsroth.easyio.EasyIOOutputStream;
import com.fernsroth.easyio.FieldHandler;
import com.fernsroth.easyio.exception.EasyIOException;
import com.fernsroth.easyio.exception.NestedEasyIOException;
import com.fernsroth.easyio.util.BeanUtils;
import com.fernsroth.easyio.util.EasyIOUtils;

/**
 * 
 * @author Joseph M. Ferner (Near Infinity Corporation)
 */
public abstract class FieldNumberHandlerBase implements FieldHandler {

    /**
     * the number of bits in this number.
     */
    private int bits;

    /**
     * true, for unsigned. false, if not unsigned.
     * TODO use this field.
     */
    @SuppressWarnings("unused")
    private boolean unsigned;

    /**
     * constructor.
     * @param bits the number of bits in this number.
     * @param unsigned true, for unsigned. false, if not unsigned.
     */
    public FieldNumberHandlerBase(int bits, boolean unsigned) {
        this.bits = bits;
        this.unsigned = unsigned;
    }

    /**
     * {@inheritDoc}
     */
    public void read(EasyIOInputStream in, Object obj, Queue<Field> fieldsQueue)
            throws EasyIOException, IOException {
        try {
            Object val = read(in);
            int bitsLeft = this.bits;
            int fieldBits = 0;
            while (true) {
                if (fieldsQueue.isEmpty()) {
                    break;
                }
                Field field = fieldsQueue.peek();
                Annotation fieldAnnotation = getAnnotation(field);
                if (fieldAnnotation == null) {
                    break;
                }
                fieldBits = getFieldBits(field, fieldAnnotation);
                if (bitsLeft < fieldBits) {
                    break;
                }
                fieldsQueue.remove();

                Object maskedVal;
                if (fieldBits == this.bits) {
                    maskedVal = val;
                } else {
                    maskedVal = EasyIOUtils.mask(val, fieldBits);
                }
                BeanUtils.setValue(obj, field, maskedVal);
                if (fieldBits != this.bits) {
                    val = EasyIOUtils.shift(val, fieldBits);
                }
                bitsLeft -= fieldBits;
            }

        } catch (IllegalArgumentException e) {
            throw new NestedEasyIOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void write(EasyIOOutputStream out, Object obj,
            Queue<Field> fieldsQueue) throws EasyIOException, IOException {
        try {
            Object val = createZero();
            int bitsLeft = this.bits;
            int fieldBits = 0;
            while (true) {
                if (fieldsQueue.isEmpty()) {
                    break;
                }
                Field field = fieldsQueue.peek();
                Annotation fieldAnnotation = getAnnotation(field);
                if (fieldAnnotation == null) {
                    break;
                }
                fieldBits = getFieldBits(field, fieldAnnotation);
                if (bitsLeft < fieldBits) {
                    break;
                }
                fieldsQueue.remove();

                Object fieldVal = BeanUtils.getValue(obj, field);
                if (fieldBits == this.bits) {
                    val = fieldVal;
                } else {
                    Object maskedVal = EasyIOUtils.mask(fieldVal, fieldBits);
                    maskedVal = EasyIOUtils.shift(maskedVal, bitsLeft
                            - this.bits);
                    val = EasyIOUtils.or(val, maskedVal);
                }
                bitsLeft -= fieldBits;
            }

            write(out, val);

        } catch (IllegalArgumentException e) {
            throw new NestedEasyIOException(e);
        }
    }

    /**
     * writes a value to the output stream.
     * @param out the output stream to write to.
     * @param val the value to write.
     * @throws IOException 
     */
    protected abstract void write(EasyIOOutputStream out, Object val)
            throws IOException;

    /**
     * creates an object with the value of zero.
     * @return the new zero object.
     */
    protected abstract Object createZero();

    /**
     * read data from the input stream.
     * @param in the input stream to read from.
     * @return the read object.
     * @throws IOException 
     */
    protected abstract Object read(EasyIOInputStream in) throws IOException;

    /**
     * gets the number of bits for a given field.
     * @param field the field to get the number of bits from/
     * @param fieldAnnotation the field annotation.
     * @return the number of bits.
     */
    protected int getFieldBits(Field field, Annotation fieldAnnotation) {
        try {
            Class<? extends Annotation> anType = fieldAnnotation
                    .annotationType();
            Method bitsMethod = anType.getMethod("bits", (Class[]) null);
            return (Integer) bitsMethod
                    .invoke(fieldAnnotation, (Object[]) null);
        } catch (IllegalArgumentException e) {
            throw new RuntimeException(fieldAnnotation.getClass().getName()
                    + " must have a 'bits' field", e);
        } catch (IllegalAccessException e) {
            throw new RuntimeException(fieldAnnotation.getClass().getName()
                    + " must have a 'bits' field", e);
        } catch (InvocationTargetException e) {
            throw new RuntimeException(fieldAnnotation.getClass().getName()
                    + " must have a 'bits' field", e);
        } catch (SecurityException e) {
            throw new RuntimeException(fieldAnnotation.getClass().getName()
                    + " must have a 'bits' field", e);
        } catch (NoSuchMethodException e) {
            throw new RuntimeException(fieldAnnotation.getClass().getName()
                    + " must have a 'bits' field", e);
        }
    }

    /**
     * @param field the field to get the annotation from.
     * @return the retrieved annotation. 
     */
    protected abstract Annotation getAnnotation(Field field);

}
