/**
 * Copyright 2006 Joseph M. Ferner
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *    
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.fernsroth.easyio.util;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import com.fernsroth.easyio.exception.GetterNotFoundException;
import com.fernsroth.easyio.exception.SetterNotFoundException;

/**
 * 
 * @author Joseph M. Ferner (Near Infinity Corporation)
 */
public final class BeanUtils {
    /**
     * hidden constructor.
     */
    private BeanUtils() {
        // empty
    }

    /**
     * gets a named value from an object.
     * @param obj the object to get the value from.
     * @param field the field to set.
     * @return the value of the named field.
     * @throws GetterNotFoundException 
     */
    public static Object getValue(Object obj, Field field)
            throws GetterNotFoundException {
        Class<?> clazz = obj.getClass();

        // get getter
        Method method = findGetterMethod(clazz, field.getName());
        try {
            if (method != null) {
                return method.invoke(obj, new Object[] {});
            } else {
                return field.get(obj);
            }
        } catch (IllegalArgumentException e) {
            throw new GetterNotFoundException(clazz, field.getName(), e);
        } catch (IllegalAccessException e) {
            throw new GetterNotFoundException(clazz, field.getName(), e);
        } catch (InvocationTargetException e) {
            throw new GetterNotFoundException(clazz, field.getName(), e);
        }
    }

    /**
     * finds a getter method in a class.
     * @param clazz the class to get the metter method from.
     * @param name the name of the field.
     * @return the getter method. null, if not found.
     */
    public static Method findGetterMethod(Class<?> clazz, String name) {
        for (Method method : clazz.getMethods()) {
            String methodName = method.getName();
            if (methodName.startsWith("get")) {
                if (methodName.substring("get".length()).equalsIgnoreCase(name)) {
                    return method;
                }
            }
        }
        return null;
    }

    /**
     * gets a list of all the fields including base classes.
     * @param clazz the class to read.
     * @return the fields.
     */
    public static List<Field> getAllFields(Class<?> clazz) {
        List<Field> fields = new ArrayList<Field>();
        Class<?> superclass = clazz.getSuperclass();
        if (superclass != null && superclass != Object.class) {
            fields.addAll(getAllFields(superclass));
        }
        fields.addAll(Arrays.asList(clazz.getDeclaredFields()));
        return fields;
    }

    /**
     * sets a value into an object.
     * @param obj the object to the value into.
     * @param field the field name.
     * @param val the value.
     * @throws SetterNotFoundException  
     */
    public static void setValue(Object obj, Field field, Object val)
            throws SetterNotFoundException {
        Class<?> clazz = obj.getClass();

        // get setter
        Method method = findSetterMethod(clazz, field.getName());
        try {
            if (method != null) {
                method.invoke(obj, new Object[] { val });
            } else {
                field.set(obj, val);
            }
        } catch (IllegalArgumentException e) {
            throw new SetterNotFoundException(clazz, field.getName(), e);
        } catch (IllegalAccessException e) {
            throw new SetterNotFoundException(clazz, field.getName(), e);
        } catch (InvocationTargetException e) {
            throw new SetterNotFoundException(clazz, field.getName(), e);
        }
    }

    /**
     * finds a setter method in the class.
     * @param clazz the class to search.
     * @param name the name of the field.
     * @return the setter method.
     */
    public static Method findSetterMethod(Class<?> clazz, String name) {
        for (Method method : clazz.getMethods()) {
            String methodName = method.getName();
            if (methodName.startsWith("set")) {
                if (methodName.substring("set".length()).equalsIgnoreCase(name)) {
                    return method;
                }
            }
        }
        return null;
    }

}
