/**
 * Copyright 2006 Joseph M. Ferner
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *    
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.fernsroth.easyio.util;

import java.io.IOException;

import com.fernsroth.easyio.CountingNullInputStream;
import com.fernsroth.easyio.EasyIOInputStream;
import com.fernsroth.easyio.exception.EasyIOException;
import com.fernsroth.easyio.exception.NestedEasyIOException;

/**
 * utilities for working with EasyRW.
 * @author Joseph M. Ferner (Near Infinity Corporation)
 */
public final class EasyIOUtils {

    /**
     * hide it. 
     */
    private EasyIOUtils() {
        // hide it.
    }

    /**
     * converts a number to a hex string.
     * @param value the value to convert.
     * @param length the length to output.
     * @return the string.
     */
    public static String toHexString(long value, int length) {
        String ret = "0000000000000000" + Long.toHexString(value);
        return ret.substring(ret.length() - length);
    }

    /**
     * gets the number of bytes the class will take up.
     * @param clazz the class.
     * @return the number of bytes.
     * @throws EasyIOException
     * @throws IOException
     */
    public static int getClassSize(Class<?> clazz) throws EasyIOException,
            IOException {
        CountingNullInputStream in = new CountingNullInputStream();
        EasyIOInputStream stream = new EasyIOInputStream(in);
        try {
            stream.read(clazz.newInstance());
        } catch (InstantiationException e) {
            throw new NestedEasyIOException(e);
        } catch (IllegalAccessException e) {
            throw new NestedEasyIOException(e);
        }
        stream.close();
        return in.getCount();
    }

    /**
     * shift the object (number) by bits.
     * @param val the value to shift.
     * @param bits the number of bits to shift.
     * @return the shifted number.
     */
    public static Object shift(Object val, int bits) {
        if (val == null) {
            throw new NullPointerException("'val' was null");
        } else if (val instanceof Long) {
            if (bits >= 0) {
                return ((Long) val).longValue() >> bits;
            } else {
                return ((Long) val).longValue() << -bits;
            }
        } else if (val instanceof Integer) {
            if (bits >= 0) {
                return ((Integer) val).intValue() >> bits;
            } else {
                return ((Integer) val).intValue() << -bits;
            }
        } else {
            throw new RuntimeException("unsupported object for shift '"
                    + val.getClass().getName() + "'");
        }
    }

    /**
     * masks an object(number) by x number of bits.
     * @param val the value to mask.
     * @param bits the number of bits to mask.
     * @return the masked number.
     */
    public static Object mask(Object val, int bits) {
        if (val == null) {
            throw new NullPointerException("'val' was null");
        }

        double mask = Math.pow(2, bits) - 1;
        if (val instanceof Long) {
            long l = ((Long) val).longValue();
            return l & (long) mask;
        } else if (val instanceof Integer) {
            int i = ((Integer) val).intValue();
            return i & (int) mask;
        }
        throw new RuntimeException("unsupported object for mask '"
                + val.getClass().getName() + "'");
    }

    /**
     * or to objects(numbers) together.
     * @param val1 the first value.
     * @param val2 the second value.
     * @return the or'ed value.
     */
    public static Object or(Object val1, Object val2) {
        if (val1 == null) {
            throw new NullPointerException("'val1' cannot be null");
        }
        if (val2 == null) {
            throw new NullPointerException("'val2' cannot be null");
        }

        if (val1 instanceof Long && val2 instanceof Long) {
            return ((Long) val1).longValue() | ((Long) val2).longValue();
        } else if (val1 instanceof Integer && val2 instanceof Long) {
            return ((Integer) val1).longValue() | ((Long) val2).longValue();
        } else if (val1 instanceof Long && val2 instanceof Integer) {
            return ((Long) val1).longValue() | ((Integer) val2).longValue();
        } else if (val1 instanceof Integer && val2 instanceof Integer) {
            return ((Integer) val1).intValue() | ((Integer) val2).intValue();
        }

        throw new RuntimeException("unsupported combination for or '"
                + val1.getClass().getName() + "' | '"
                + val2.getClass().getName() + "'");
    }
}
